import pytest

from bocadillo import App, HTTPError


def test_if_no_route_exists_for_name_then_url_for_raises_404(app: App):
    with pytest.raises(HTTPError):
        app.url_for(name="about")


def test_if_route_exists_then_url_for_returns_full_path(app: App):
    @app.route("/about/{who}")
    async def about(req, res, who):
        pass

    url = app.url_for("about", who="me")
    assert url == "/about/me"


def test_name_can_be_explicitly_given(app: App):
    @app.route("/about/{who}", name="about-someone")
    async def about(req, res, who):
        pass

    with pytest.raises(HTTPError):
        app.url_for(name="about", who="me")

    url = app.url_for("about-someone", who="me")
    assert url == "/about/me"


def test_name_is_inferred_from_view_name(app: App):
    @app.route("/about/{who}")
    class AboutPerson:
        async def get(self, req, res, who):
            pass

    url = app.url_for("about_person", who="Godzilla")
    assert url == "/about/Godzilla"

    @app.route("/about/{who}")
    async def about_who(req, res, who):
        pass

    url = app.url_for("about_who", who="Godzilla")
    assert url == "/about/Godzilla"


def test_if_route_has_namespace_then_must_be_used_when_reversing(app: App):
    @app.route("/about", namespace="blog")
    async def about(req, res):
        pass

    assert app.url_for("blog:about") == "/about"

    with pytest.raises(HTTPError):
        app.url_for("about")


def test_reverse_named_sub_app_route(app: App):
    sub = App("sub")

    @sub.route("/foo")
    async def foo(req, res):
        pass

    app.mount("/sub", sub)

    assert app.url_for("sub:foo") == "/sub/foo"


def test_cannot_reverse_unnamed_sub_app_route(app: App):
    sub = App()

    @sub.route("/foo")
    async def foo(req, res):
        pass

    app.mount("/sub", sub)

    with pytest.raises(HTTPError):
        app.url_for("sub:foo")



import os
import random
from multiprocessing import Event, Process
from typing import TYPE_CHECKING

from starlette.testclient import TestClient

try:
    import pytest
except ImportError:
    pytest = None

if TYPE_CHECKING:
    from .applications import App


def create_client(app: "App", **kwargs) -> TestClient:
    """Create a [Starlette Test Client][client] out of an application.

    [client]: https://www.starlette.io/testclient/

    # Parameters
    app: an #::bocadillo.applications#App instance.
    **kwargs (any): keyword arguments passed to the `TestClient` constructor.

    # Returns
    client (TestClient): a Starlette test client.
    """
    return TestClient(app, **kwargs)


class LiveServer:
    """Context manager to spin up a live server in a separate process.

    The server process is terminated when exiting the context.

    # Parameters
    app: an #::bocadillo.applications#App instance.
    host (str):
        the host to run the server on.
        Defaults to `"127.0.0.1."` (i.e. `localhost`).
    port (int):
        the port to run the server on.
        Defaults to a random integer between 3000 and 9000.
    ready_timeout (float):
        The maximum time to wait for the live server to be ready to handle
        connections, in seconds.
        Defaults to 5.
    stop_timeout (float):
        The maximum time to wait for the live server to terminate, in seconds.
        Defaults to 5.
    
    # Attributes
    url (str):
        the full URL where the server lives.
    """

    def __init__(
        self,
        app: "App",
        host: str = "127.0.0.1",
        port: int = None,
        ready_timeout: float = 5,
        stop_timeout: float = 5,
    ):
        if (
            pytest is not None and os.getenv("CI") and os.getenv("TRAVIS")
        ):  # pragma: no cover
            pytest.skip("live server process sometimes makes travis ci stall")

        if port is None:
            port = random.randint(3000, 9000)

        self.app = app
        self.host = host
        self.port = port
        self.ready_timeout = ready_timeout
        self.stop_timeout = stop_timeout
        self._process: Process = None

    @property
    def url(self):
        return f"http://{self.host}:{self.port}"

    def __enter__(self):
        ready = Event()

        def target():
            async def callback_notify():
                # Run periodically by the Uvicorn server.
                ready.set()

            self.app.run(
                host=self.host, port=self.port, callback_notify=callback_notify
            )

        self._process = Process(target=target)
        self._process.start()

        if not ready.wait(self.ready_timeout):  # pragma: no cover
            raise TimeoutError(
                f"Live server not ready after {self.ready_timeout} seconds"
            )

        return self

    def __exit__(self, *args):
        self._process.terminate()
        self._process.join(self.stop_timeout)
        if self._process.exitcode is None:  # pragma: no cover
            raise TimeoutError(
                f"Live server still running after {self.stop_timeout} seconds."
            )



from asyncio import sleep
from time import sleep as sync_sleep
from multiprocessing import Value

import pytest
import requests

from bocadillo import App, ClientDisconnect
from bocadillo.testing import LiveServer

from .utils import stops_incrementing


def test_stream_response(app: App, client):
    background_called = False

    @app.route("/{word}")
    async def index(req, res, word: str):
        @res.stream
        async def stream_word():
            for character in word:
                yield character
                await sleep(0.001)

        # Other features of res can also be used

        res.headers["x-foo"] = "foo"
        res.status_code = 202

        @res.background
        async def do_later():
            nonlocal background_called
            background_called = True

    r = client.get("/hello")
    assert r.text == "hello"
    assert r.headers["x-foo"] == "foo"
    assert r.status_code == 202
    assert background_called
    # streaming does not send the response in chunks
    assert "transfer-encoding" not in r.headers


def test_stream_func_must_be_async_generator_function(app: App):
    @app.route("/")
    async def index(req, res):
        with pytest.raises(AssertionError):
            # Regular function
            @res.stream
            def foo():
                pass

        with pytest.raises(AssertionError):
            # Coroutine function
            @res.stream
            async def foo():
                pass

        with pytest.raises(AssertionError):
            # Regular generator
            @res.stream
            def foo():
                yield "nope"


def test_stop_on_client_disconnect(app: App):
    sent = Value("i", 0)

    @app.route("/inf")
    async def infinity(req, res):
        @res.stream
        async def stream():
            nonlocal sent
            while True:
                yield "∞"
                sent.value += 1

    with LiveServer(app) as server:
        r = requests.get(f"{server.url}/inf", stream=True)
        assert r.status_code == 200
        assert stops_incrementing(counter=sent, response=r)


def test_raise_on_disconnect(app: App):
    caught = Value("i", 0)

    @app.route("/inf")
    async def infinity(req, res):
        @res.stream(raise_on_disconnect=True)
        async def stream():
            nonlocal caught
            try:
                while True:
                    yield "∞"
            except ClientDisconnect:
                caught.value = 1

    with LiveServer(app) as server:
        r = requests.get(f"{server.url}/inf", stream=True)
        assert r.status_code == 200
        r.close()
        sync_sleep(0.1)
        assert caught.value



from typing import NamedTuple

import pytest

from bocadillo import App, Templates, Recipe
from bocadillo.testing import create_client


APP_CLASSES = [App, lambda: Recipe("tacos")]
CLIENT_FACTORIES = [create_client, lambda app: app.client]


@pytest.fixture(params=APP_CLASSES, name="app")
def fixture_app(request):
    cls = request.param
    return cls()


@pytest.fixture(params=CLIENT_FACTORIES)
def client(app, request):
    factory = request.param
    return factory(app)


@pytest.fixture(name="templates")
def fixture_templates(app: App):
    return Templates(app)


class TemplateWrapper(NamedTuple):
    name: str
    context: dict
    rendered: str
    root: str


def create_template(
    templates: Templates, tmpdir_factory, dirname: str
) -> TemplateWrapper:
    templates_dir = tmpdir_factory.mktemp(dirname)

    template = templates_dir.join("hello.html")
    template.write("<h1>Hello, {{ name }}!</h1>")

    templates.directory = str(templates_dir)

    return TemplateWrapper(
        name="hello.html",
        context={"name": "Bocadillo"},
        rendered="<h1>Hello, Bocadillo!</h1>",
        root=str(templates_dir),
    )


@pytest.fixture
def template_file(templates: Templates, tmpdir_factory) -> TemplateWrapper:
    return create_template(templates, tmpdir_factory, dirname="templates")



import pytest

from bocadillo import App


def test_root_wildcard(app: App, client):
    @app.route("{}")
    async def root(req, res):
        res.text = "root"

    r = client.get("/")
    assert r.status_code == 200
    assert r.text == "root"


def test_wildcard(app: App, client):
    @app.route("/foo/{}-bar")
    async def foo_bar(req, res):
        # Field has no name => not passed as a parameter to view.
        # NOTE: if user wants to retrieve the value of the matched field,
        # they must give it a name.
        pass

    r = client.get("/foo/no-bar")
    assert r.status_code == 200


def test_if_wildcard_first_then_other_routes_wont_match(app: App, client):
    @app.route("/foo/{}")
    async def foo_all(req, res):
        res.text = "all"

    @app.route("/foo/bar")
    async def foo_bar(req, res):
        res.text = "bar"

    r = client.get("/foo/bar")
    assert r.status_code == 200
    assert r.text == "all"


@pytest.mark.parametrize(
    "path, text", [("/foo/bar", "bar"), ("/foo/jong", "all")]
)
def test_if_wildcard_last_then_matches_by_default(
    app: App, client, path: str, text: str
):
    @app.route("/foo/bar")
    async def foo_bar(req, res):
        res.text = "bar"

    @app.route("/foo/{}")
    async def foo_all(req, res):
        res.text = "all"

    r = client.get(path)
    assert r.status_code == 200
    assert r.text == text



import warnings
from functools import partial, wraps
from inspect import cleandoc, isclass
from typing import Callable, Tuple, Type, TypeVar, Union, cast

_F = TypeVar("_F", bound=Callable)
_T = TypeVar("_T")
_FT = Union[_F, Type[_T]]


def deprecated(
    since: str,
    removal: str,
    alternative: Union[str, Tuple[str, str]],
    update_doc: bool = True,
    warn_on_instanciate: bool = False,
) -> Callable:
    """Mark a function or a class as deprecated.

    The function or class will raise a `DeprecationWarning` when called and
    its docstring will be updated with a warning special container to be
    rendered in the API reference.

    # Parameters
    since (str):
        version when the function has been deprecated.
    removal (str):
        version when the function will be removed.
    alternative (str or tuple):
        what should be used instead. Can be a string (name of the alternative)
        or a tuple of two strings (name and a link to the API reference).
    link (str):
        a link to the API reference for the alternative.
    update_doc (bool):
        whether the object's docstring should be updated.
    warn_on_instanciate (bool):
        if the object is a class, whether a deprecation warning should be
        sent when instanciating it.    
    """

    if isinstance(alternative, tuple):
        name, link = alternative
        alternative_formatted = f"[`{name}`]({link})"
        alternative = name
    else:
        alternative_formatted = f"`{alternative}`"

    def get_message(obj: _FT, strip: bool = False) -> str:
        obj_name = obj.__name__ if strip else f"`{obj.__name__}`"
        alt = alternative if strip else alternative_formatted
        return (
            f"{obj_name} was **deprecated** in v{since}, and will "
            f"be **removed** in v{removal}. Please use {alt} instead."
        )

    def show_warning(obj: _FT):
        warnings.simplefilter("always", DeprecationWarning)
        warnings.warn(
            get_message(obj, strip=True),
            category=DeprecationWarning,
            stacklevel=2,
        )
        warnings.simplefilter("default", DeprecationWarning)

    def get_doc_warning(obj: _FT) -> str:
        return f"::: warning DEPRECATED\n{get_message(obj)}\n:::"

    def add_warning(obj: _FT) -> _FT:
        if isclass(obj):
            cls = cast(Type, obj)

            if warn_on_instanciate:

                class wrapped(cls):  # type: ignore
                    def __init__(self, *args, **kwargs):
                        show_warning(obj)
                        super().__init__(*args, **kwargs)

                # No equivalent for @wraps for classes? Update name and docs.
                wrapped.__name__ = cls.__name__
                wrapped.__doc__ = cls.__doc__

            else:
                wrapped = cls  # type: ignore

        else:
            func = cast(Callable, obj)

            @wraps(func)  # type: ignore
            def wrapped(*args, **kwargs):
                show_warning(func)
                return func(*args, **kwargs)

        if update_doc:
            header, _, body = cleandoc(obj.__doc__ or "").partition("\n")
            wrapped.__doc__ = "\n\n".join([header, get_doc_warning(obj), body])

        return wrapped

    return add_warning


class ReplacedBy:
    def __init__(
        self, since: str, removal: str, obj_root: str = "", doc_root: str = ""
    ):
        self._obj_root = obj_root
        self._doc_root = doc_root
        self._factory = partial(deprecated, since=since, removal=removal)

    def __call__(self, name: str, fragment: str = None):
        alternative: Union[str, tuple]
        if fragment is not None:
            alternative = (self._obj_root + name, self._doc_root + fragment)
        else:
            alternative = self._obj_root + name
        return self._factory(alternative=alternative)



import pytest

from bocadillo import App
from bocadillo.testing import create_client


def test_false_by_default(app: App):
    assert app.debug is False


def test_setter_updates_exception_and_error_middleware(app: App):
    app.debug = True
    assert app.debug is True
    assert app.exception_middleware.debug is True
    assert app.server_error_middleware.debug is True


def test_if_not_debug_then_no_debug_info_returned(app: App):
    @app.route("/")
    async def index(req, res):
        raise ValueError("Oops")

    client = create_client(app, raise_server_exceptions=False)
    r = client.get("/")
    assert r.status_code == 500
    assert r.text == "500 Internal Server Error"


@pytest.mark.parametrize(
    "accept, content_type",
    [
        (None, "text/plain; charset=utf-8"),
        ("text/html", "text/html; charset=utf-8"),
    ],
)
def test_debug_response(app: App, accept: str, content_type: str):
    app.debug = True

    @app.route("/")
    async def index(req, res):
        raise ValueError("Oops")

    client = create_client(app, raise_server_exceptions=False)
    r = client.get("/", headers={"accept": accept})
    assert r.status_code == 500
    assert r.headers["content-type"] == content_type
    assert 'raise ValueError("Oops")' in r.text



import traceback
from http import HTTPStatus
from typing import Any, Dict, Optional, Type, Union

import jinja2
from starlette.responses import HTMLResponse, PlainTextResponse

from .app_types import _E, ErrorHandler, HTTPApp
from .compat import call_async
from .misc import read_asset
from .request import Request
from .response import Response


class HTTPError(Exception):
    """Raised when an HTTP error occurs.

    You can raise this within a view or an error handler to interrupt
    request processing.

    # Parameters
    status (int or HTTPStatus):
        the status code of the error.
    detail (any):
        extra detail information about the error. The exact rendering is
        determined by the configured error handler for `HTTPError`.

    # See Also
    - [HTTP response status codes (MDN web docs)](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status)
    """

    def __init__(self, status: Union[int, HTTPStatus], detail: Any = ""):
        if isinstance(status, int):
            status = HTTPStatus(  # pylint: disable=no-value-for-parameter
                status
            )
        else:
            assert isinstance(
                status, HTTPStatus
            ), f"Expected int or HTTPStatus, got {type(status)}"
        self._status = status
        self.detail = detail

    @property
    def status_code(self) -> int:
        """Return the HTTP error's status code, e.g. `404`."""
        return self._status.value

    @property
    def status_phrase(self) -> str:
        """Return the HTTP error's status phrase, e.g. `"Not Found"`."""
        return self._status.phrase

    @property
    def title(self) -> str:
        """Return the HTTP error's title, e.g. `"404 Not Found"`."""
        return f"{self.status_code} {self.status_phrase}"

    def __str__(self):
        return self.title


class ServerErrorMiddleware(HTTPApp):
    """Return 500 response when an unhandled exception occurs.

    Adaptation of Starlette's `ServerErrorMiddleware`.
    """

    _template_name = "server_error.jinja"

    def __init__(
        self, app: HTTPApp, handler: ErrorHandler, debug: bool = False
    ) -> None:
        self.app = app
        self.handler = handler
        self.debug = debug
        self.exception: Optional[BaseException] = None
        self.jinja = jinja2.Environment()

    def generate_html(self, req: Request, exc: BaseException) -> str:
        template = self.jinja.from_string(read_asset(self._template_name))
        tb_exc = traceback.TracebackException.from_exception(
            exc, capture_locals=True
        )
        return template.render(
            exc_type=exc.__class__.__name__,
            exc=exc,
            url_path=req.url.path,
            frames=tb_exc.stack,
        )

    def generate_plain_text(self, exc: BaseException) -> str:
        return "".join(traceback.format_tb(exc.__traceback__))

    def debug_response(self, req: Request, exc: BaseException) -> Response:
        accept = req.headers.get("accept", "")

        if "text/html" in accept:
            content = self.generate_html(req, exc)
            return HTMLResponse(content, status_code=500)

        content = self.generate_plain_text(exc)
        return PlainTextResponse(content, status_code=500)

    def raise_if_exception(self):
        if self.exception is not None:
            raise self.exception from None

    async def __call__(self, req: Request, res: Response) -> Response:
        try:
            res = await self.app(req, res)
        except BaseException as exc:
            self.exception = exc
            if self.debug:
                # In debug mode, return traceback responses.
                res = self.debug_response(req, exc)
            await call_async(  # type: ignore
                self.handler, req, res, HTTPError(500)
            )
            return res
        else:
            return res


class HTTPErrorMiddleware(HTTPApp):
    """Handle exceptions that occur while handling HTTP requests.

    Adaptation of Starlette's `ExceptionMiddleware`.
    """

    def __init__(self, app: HTTPApp, debug: bool = False) -> None:
        self.app = app
        self.debug = debug
        self._exception_handlers: Dict[Type[BaseException], ErrorHandler] = {}

    def add_exception_handler(
        self, exception_class: Type[_E], handler: ErrorHandler
    ) -> None:
        assert issubclass(exception_class, BaseException)
        self._exception_handlers[exception_class] = handler

    def _get_exception_handler(self, exc: _E) -> Optional[ErrorHandler]:
        for cls, handler in self._exception_handlers.items():
            if issubclass(type(exc), cls):
                return handler
        return None

    async def __call__(self, req: Request, res: Response) -> Response:
        try:
            res = await self.app(req, res)
        except Exception as exc:  # pylint: disable=broad-except
            handler = self._get_exception_handler(exc)
            if handler is None:
                raise exc from None
            await call_async(handler, req, res, exc)  # type: ignore
            return res
        else:
            return res



from typing import AnyStr, Callable, AsyncIterable
import inspect

from .request import Request, ClientDisconnect

Stream = AsyncIterable[AnyStr]
StreamFunc = Callable[[], Stream]


def stream_until_disconnect(
    req: Request, source: Stream, raise_on_disconnect: bool
) -> Stream:
    # Yield items from a stream until the client disconnects, then
    # throw an exception into the stream (if told to do so).

    assert inspect.isasyncgen(source)

    async def stream():
        async for item in source:
            if not await req.is_disconnected():
                yield item
                continue

            if raise_on_disconnect:
                try:
                    await source.athrow(ClientDisconnect)
                except StopAsyncIteration:
                    # May be raised in Python 3.6 if the `source`'s error
                    # handling code did not `yield` anything.
                    break
            else:
                break

    return stream()



from .applications import App
from .errors import HTTPError
from .middleware import ASGIMiddleware, Middleware
from .injection import discover_providers, provider, useprovider
from .recipes import Recipe
from .request import ClientDisconnect, Request
from .response import Response
from .sse import server_event
from .staticfiles import static
from .templates import Templates
from .views import view
from .websockets import WebSocket, WebSocketDisconnect

__version__ = "0.13.3"



from bocadillo import App
from bocadillo.testing import create_client


def test_if_host_not_allowed_then_400():
    app = App(allowed_hosts=["example.com"])
    client = create_client(app)

    @app.route("/")
    async def index(req, res):
        pass

    response = client.get("/")
    assert response.status_code == 400



from http import HTTPStatus

import pytest

from bocadillo import App, HTTPError
from bocadillo.error_handlers import (
    error_to_html,
    error_to_media,
    error_to_text,
)
from bocadillo.testing import create_client


@pytest.mark.parametrize(
    "status",
    [
        "400 Bad Request",
        "401 Unauthorized",
        "403 Forbidden",
        "405 Method Not Allowed",
        "500 Internal Server Error",
        # Non-error codes are supported too. Be responsible.
        "200 OK",
        "201 Created",
        "202 Accepted",
        "204 No Content",
    ],
)
def test_if_http_error_is_raised_then_automatic_response_is_sent(
    app: App, client, status: str
):
    status, phrase = status.split(" ", 1)
    status_code = int(status)

    @app.route("/")
    async def index(req, res):
        raise HTTPError(status_code)

    response = client.get("/")
    assert response.status_code == status_code
    assert phrase in response.text


@pytest.mark.parametrize(
    "exception_cls", [KeyError, ValueError, AttributeError]
)
def test_custom_error_handler(app: App, exception_cls):
    called = False

    @app.error_handler(KeyError)
    def on_key_error(req, res, exc):
        nonlocal called
        res.text = "Oops!"
        called = True

    @app.route("/")
    async def index(req, res):
        raise exception_cls("foo")

    client = create_client(app, raise_server_exceptions=False)

    if exception_cls == KeyError:
        response = client.get("/")
        assert called
        assert response.status_code == 200
        assert response.text == "Oops!"
    else:
        response = client.get("/")
        assert response.status_code == 500
        assert not called


# Use in a test to run against multiple error details. See:
# https://docs.pytest.org/en/latest/fixture.html#parametrizing-fixtures
@pytest.fixture(params=["", "Nope!"])
def detail(request):
    return request.param


@pytest.mark.parametrize(
    "handler, content, expected",
    [
        (
            error_to_html,
            lambda res: res.text,
            lambda detail: "<h1>403 Forbidden</h1>{}".format(
                f"\n<p>{detail}</p>" if detail else ""
            ),
        ),
        (
            error_to_media,
            lambda res: res.json(),
            lambda detail: (
                {"error": "403 Forbidden", "detail": detail, "status": 403}
                if detail
                else {"error": "403 Forbidden", "status": 403}
            ),
        ),
        (
            error_to_text,
            lambda res: res.text,
            lambda detail: "403 Forbidden{}".format(
                f"\n{detail}" if detail else ""
            ),
        ),
    ],
)
def test_builtin_handlers(app: App, client, detail, handler, content, expected):
    app.add_error_handler(HTTPError, handler)

    @app.route("/")
    async def index(req, res):
        raise HTTPError(403, detail=detail)

    response = client.get("/")
    assert response.status_code == 403
    assert content(response) == expected(detail)


def test_http_error_status_must_be_int_or_http_status():
    HTTPError(404)
    HTTPError(HTTPStatus.NOT_FOUND)

    with pytest.raises(AssertionError) as ctx:
        HTTPError("404")

    assert "int or HTTPStatus" in str(ctx.value)


def test_http_error_str_representation():
    assert str(HTTPError(404, detail="foo")) == "404 Not Found"



from typing import Any, Awaitable, Callable, MutableMapping, Union, TypeVar

from .request import Request
from .response import Response

_E = TypeVar("_E", bound=Exception)

# ASGI
Scope = dict
Event = MutableMapping[str, Any]
Receive = Callable[[], Awaitable[Event]]
Send = Callable[[Event], None]


class ASGIAppInstance:
    def __call__(self, receive: Receive, send: Send) -> None:
        raise NotImplementedError


class ASGIApp:
    def __call__(self, scope: Scope) -> ASGIAppInstance:
        raise NotImplementedError


# HTTP
AsyncHandler = Callable[[Request, Response, Any], Awaitable[None]]
SyncHandler = Callable[[Request, Response, Any], None]
Handler = Union[AsyncHandler, SyncHandler]
ErrorHandler = Callable[[Request, Response, _E], Awaitable[None]]


class HTTPApp:
    async def __call__(self, req: Request, res: Response) -> Response:
        raise NotImplementedError


# Server lifespan events
EventHandler = Callable[[], Awaitable[None]]



from bocadillo import App
from bocadillo.testing import create_client


def test_no_allowed_origins_by_default():
    app = App(enable_cors=True)

    @app.route("/")
    async def index(req, res):
        pass

    client = create_client(app)
    response = client.options(
        "/",
        headers={
            "origin": "foobar.com",
            "access-control-request-method": "GET",
        },
    )
    assert response.status_code == 400


def test_if_origin_not_in_allow_origins_then_400():
    app = App(enable_cors=True, cors_config={"allow_origins": ["foobar.com"]})

    @app.route("/")
    async def index(req, res):
        pass

    client = create_client(app)
    response = client.options(
        "/",
        headers={
            "origin": "foobar.com",
            "access-control-request-method": "GET",
        },
    )
    assert response.status_code == 200

    response = client.options(
        "/",
        headers={
            "origin": "example.com",
            "access-control-request-method": "GET",
        },
    )
    assert response.status_code == 400


def test_if_method_not_in_allow_methods_then_400():
    app = App(
        enable_cors=True,
        cors_config={
            "allow_origins": ["foobar.com"],
            "allow_methods": ["POST"],
        },
    )

    @app.route("/")
    async def index(req, res):
        pass

    client = create_client(app)
    response = client.options(
        "/",
        headers={
            "origin": "foobar.com",
            "access-control-request-method": "GET",
        },
    )
    assert response.status_code == 400



from os.path import join, dirname, abspath

_ASSETS_DIR = join(dirname(abspath(__file__)), "assets")


# TODO make async using aiofiles?
def read_asset(filename: str) -> str:
    with open(join(_ASSETS_DIR, filename), "r") as f:
        return f.read()



import pytest

from bocadillo import App, view
from bocadillo.compat import nullcontext
from bocadillo.sessions import MissingSecretKey
from bocadillo.testing import create_client
from bocadillo.utils import override_env


def test_sessions_enabled_no_secret_key():
    with pytest.raises(MissingSecretKey):
        App(enable_sessions=True)


def test_sessions_enabled_secret_key_empty():
    with override_env("SECRET_KEY", ""):
        with pytest.raises(MissingSecretKey):
            App(enable_sessions=True)


@pytest.mark.parametrize(
    "ctx, config",
    (
        [override_env("SECRET_KEY", "not-so-secret"), {}],
        [nullcontext(), {"secret_key": "not-so-secret"}],
    ),
)
def test_sessions_enabled_secret_key_present(ctx, config):
    with ctx:
        app = App(enable_sessions=True, sessions_config=config)

        @app.route("/set")
        @view(methods=["post"])
        async def set_session(req, res):
            req.session["data"] = "something"
            res.text = "Saved"

        @app.route("/")
        async def index(req, res):
            data = req.session["data"]
            res.text = f"Hello {data}"

        client = create_client(app)
        client.post("/set")
        response = client.get("/")
        assert "something" in response.text
        assert "session" in response.cookies



import pytest

from bocadillo import App, provider, WebSocket


@pytest.fixture(autouse=True)
def clients_provider():
    @provider(scope="app")
    async def clients():
        return set()


def test_websocket_clients_example(app: App, client):
    @app.websocket_route("/chat")
    async def chat(ws: WebSocket, clients):
        clients.add(ws)
        await ws.send_text(str(len(clients)))

    @app.route("/clients")
    async def client_count(req, res, clients):
        res.media = {"count": len(clients)}

    with client.websocket_connect("/chat") as ws:
        assert ws.receive_text() == "1"

    r = client.get("/clients")
    assert r.status_code == 200
    assert r.json() == {"count": 1}



import json

import pytest

from bocadillo import App
from bocadillo.constants import CONTENT_TYPE
from bocadillo.media import UnsupportedMediaType


def test_defaults_to_json(app: App, client):
    data = {"message": "hello"}

    @app.route("/")
    async def index(req, res):
        res.media = data

    response = client.get("/")
    assert response.status_code == 200
    assert response.headers["content-type"] == "application/json"
    assert response.json() == data


def test_can_specify_media_type_when_creating_the_api_object():
    App(media_type=CONTENT_TYPE.JSON)


def test_media_type_is_accessible_on_api(app: App):
    assert hasattr(app, "media_type")


@pytest.mark.parametrize(
    "media_type, expected_text", [(CONTENT_TYPE.JSON, json.dumps)]
)
def test_use_builtin_media_handlers(
    app: App, client, media_type: str, expected_text
):
    app.media_type = media_type
    data = {"message": "hello"}

    @app.route("/")
    async def index(req, res):
        res.media = data

    response = client.get("/")
    assert response.status_code == 200
    assert response.headers["content-type"] == media_type
    assert response.text == expected_text(data)


@pytest.fixture
def foo_type():
    return "application/foo"


@pytest.fixture
def handle_foo():
    return lambda value: f"FOO: {value}"


def test_add_and_use_custom_media_handler(
    app: App, client, foo_type, handle_foo
):
    app.media_handlers[foo_type] = handle_foo
    app.media_type = foo_type

    @app.route("/")
    async def index(req, res):
        res.media = "bar"

    response = client.get("/")
    assert response.status_code == 200
    assert response.headers["content-type"] == foo_type
    assert response.text == handle_foo("bar")


def test_replace_media_handlers(app: App, client, foo_type, handle_foo):
    app.media_handlers = {foo_type: handle_foo}
    app.media_type = foo_type

    @app.route("/")
    async def index(req, res):
        res.media = "bar"

    response = client.get("/")
    assert response.status_code == 200
    assert response.headers["content-type"] == foo_type
    assert response.text == handle_foo("bar")


def test_if_media_type_not_supported_then_passing_it_raises_error(
    app: App, foo_type
):
    with pytest.raises(UnsupportedMediaType) as ctx:
        App(media_type=foo_type)

    assert foo_type in str(ctx.value)


def test_if_media_type_not_supported_then_setting_it_raises_error(
    app: App, foo_type
):
    with pytest.raises(UnsupportedMediaType) as ctx:
        app.media_type = foo_type

    assert foo_type in str(ctx.value)



import pytest

from bocadillo import App, provider


@pytest.fixture(autouse=True)
async def hello_provider():
    @provider
    async def hello_format():
        return "Hello, {who}!"

    @provider
    async def hello(hello_format) -> str:
        return hello_format.format(who="providers")


def test_function_based_view(app: App, client):
    @app.route("/hi")
    async def say_hi(req, res, hello):
        res.text = hello

    r = client.get("/hi")
    assert r.status_code == 200
    assert r.text == "Hello, providers!"


def test_function_based_view_with_route_parameters(app: App, client):
    @app.route("/hi/{who}")
    async def say_hi(req, res, hello_format, who):
        res.text = hello_format.format(who=who)

    r = client.get("/hi/peeps")
    assert r.status_code == 200
    assert r.text == "Hello, peeps!"


def test_class_based_view(app: App, client):
    @app.route("/hi")
    class SayHi:
        async def get(self, req, res, hello):
            res.text = hello

    r = client.get("/hi")
    assert r.status_code == 200
    assert r.text == "Hello, providers!"


def test_class_based_view_with_route_parameters(app: App, client):
    @app.route("/hi/{who}")
    class SayHi:
        async def get(self, req, res, hello_format, who):
            res.text = hello_format.format(who=who)

    r = client.get("/hi/peeps")
    assert r.status_code == 200
    assert r.text == "Hello, peeps!"



import pytest

from bocadillo import App


def _setup_views_with_redirect(app, permanent: bool = None, **kwargs):
    @app.route("/home")
    async def home(req, res):
        res.text = "You are home!"

    @app.route("/")
    async def index(req, res):
        if permanent:
            app.redirect(permanent=True, **kwargs)
        else:
            app.redirect(**kwargs)


def test_redirect_by_route_name(app: App, client):
    _setup_views_with_redirect(app, name="home")
    response = client.get("/")
    assert response.status_code == 200
    assert response.text == "You are home!"


def test_if_redirect_to_non_existing_route_then_404(app: App, client):
    _setup_views_with_redirect(app, name="about")
    response = client.get("/")
    assert response.status_code == 404


def test_redirection_is_temporary_by_default(app: App, client):
    _setup_views_with_redirect(app, name="home")
    response = client.get("/", allow_redirects=False)
    assert response.status_code == 302


def test_permanent_redirect(app: App, client):
    _setup_views_with_redirect(app, permanent=True, name="home")
    response = client.get("/", allow_redirects=False)
    assert response.status_code == 301


def test_at_least_one_of_name_or_url_must_be_given(app: App):
    with pytest.raises(AssertionError) as ctx:
        app.redirect()
    assert all(item in str(ctx.value) for item in ("url", "expected", "name"))


def test_redirect_to_internal_url(app: App, client):
    @app.route("/about/{who}")
    async def about(req, res, who):
        res.text = who

    @app.route("/")
    async def index(req, res):
        app.redirect(url="/about/me")

    response = client.get("/")
    assert response.status_code == 200
    assert response.text == "me"


def test_if_redirect_to_non_matching_internal_url_then_404(app: App, client):
    @app.route("/")
    async def index(req, res):
        app.redirect(url="/about/me")

    response = client.get("/")
    assert response.status_code == 404


def test_redirect_to_external_url(app: App, client):
    external_url = "https://httpbin.org/status/202"

    @app.route("/")
    async def index(req, res):
        app.redirect(url=external_url)

    # NOTE: cannot follow redirect here, because the TestClient
    # prepends the base_url (http://testserver) to any request made, including
    # the redirected request.
    response = client.get("/", allow_redirects=False)
    assert response.status_code == 302
    assert response.headers["location"] == external_url



import pytest

from bocadillo import App, ASGIMiddleware
from bocadillo.testing import create_client


def test_asgi_middleware(app: App, client):
    params = None
    received_app = False
    called = False

    class Middleware(ASGIMiddleware):
        def __init__(self, inner, app: App, **kwargs):
            super().__init__(inner, app)
            nonlocal params, received_app
            params = kwargs
            received_app = isinstance(app, App)

        def __call__(self, scope):
            nonlocal called
            called = True
            return super().__call__(scope)

    app.add_asgi_middleware(Middleware, hello="world")
    assert received_app
    assert params == {"hello": "world"}

    @app.route("/")
    async def index(req, res):
        pass

    client.get("/")
    assert called


def test_pure_asgi_middleware(app: App, client):
    initialized = False
    called = False

    class Middleware:
        def __init__(self, inner):
            nonlocal initialized
            self.inner = inner
            initialized = True

        def __call__(self, scope: dict):
            nonlocal called
            called = True
            return self.inner(scope)

    app.add_asgi_middleware(Middleware)

    assert initialized

    @app.route("/")
    async def index(req, res):
        pass

    client.get("/")
    assert called


@pytest.mark.parametrize(
    ["route", "origin", "expected", "expected_body"],
    [
        ("/", "localhost:8001", "localhost:8001", "Hello"),
        ("/", "ietf.org", "ietf.org", "Hello"),
        ("/", "unknown.org", None, "Hello"),
        ("/sub/", "localhost:8001", "localhost:8001", "OK"),
        ("/sub/", "ietf.org", "ietf.org", "OK"),
        ("/sub/", "unknown.org", None, "OK"),
    ],
)
def test_middleware_called_if_routed_to_sub_app(
    route: str, origin: str, expected: str, expected_body: str
):
    app = App(
        enable_cors=True,
        cors_config={
            "allow_origins": ["example.com", "localhost:8001", "ietf.org"]
        },
    )

    @app.route("/")
    async def index(req, res):
        res.text = "Hello"

    sub = App()

    @sub.route("/")
    class SubApp:
        async def get(self, req, res):
            res.text = "OK"

    app.mount("/sub", sub)

    client = create_client(app)
    res = client.get(route, headers={"origin": origin})

    assert res.text == expected_body

    if not expected:  # unknown origin -> no allow-origin header
        assert "access-control-allow-origin" not in res.headers
    else:  # allowed origin -> allow-origin header"
        assert "access-control-allow-origin" in res.headers
        assert res.headers.get("access-control-allow-origin") == expected



import pytest

from bocadillo import App, Recipe


@pytest.fixture
def numbers():
    integers = Recipe("integers")

    @integers.route("/{x}")
    async def convert(req, res, x):
        res.media = {"value": int(float(x))}

    floats = Recipe("floats")

    @floats.route("/{x}")
    async def convert(req, res, x):
        res.media = {"value": float(x)}

    return Recipe.book(integers, floats, prefix="/numbers")


def test_recipe_book(app: App, client, numbers):
    app.recipe(numbers)

    response = client.get("/numbers/integers/3.5")
    assert response.status_code == 200
    assert response.json() == {"value": 3}

    response = client.get("/numbers/floats/1")
    assert response.status_code == 200
    assert response.json() == {"value": 1.0}



from bocadillo import App


def test_attachment(app: App, client):
    @app.route("/")
    async def index(req, res):
        res.text = "hello.text"
        res.attachment = "hello.txt"

    response = client.get("/")
    assert response.status_code == 200
    assert (
        response.headers["content-disposition"]
        == "attachment; filename='hello.txt'"
    )



from bocadillo import App


def test_parameter_is_passed_as_argument(app: App, client):
    @app.route("/greet/{person}")
    async def greet(req, res, person: str):
        res.text = person

    response = client.get("/greet/John")
    assert response.status_code == 200
    assert response.text == "John"


def test_if_route_expects_int_but_int_not_given_then_404(app: App, client):
    @app.route("/add/{x:d}/{y:d}")
    async def add(req, res, x, y):
        res.text = str(x + y)

    response = client.get("/add/1/foo")
    assert response.status_code == 404



from contextlib import suppress

import pytest

from bocadillo import WebSocket, App, WebSocketDisconnect
from bocadillo.constants import WEBSOCKET_CLOSE_CODES


# Basic usage


def test_websocket_route(app: App, client):
    @app.websocket_route("/chat")
    async def chat(ws: WebSocket):
        assert await ws.receive_text() == "ping"
        await ws.send_text("pong")

    with client.websocket_connect("/chat") as ws:
        ws.send_text("ping")
        assert ws.receive_text() == "pong"


def test_enter_context_twice_is_safe(app: App, client):
    # Ensures compatibility with <0.13.1.

    @app.websocket_route("/chat")  # auto_accept==True
    async def chat(ws: WebSocket):
        async with ws:  # <-- should be OK (previous API)
            assert await ws.receive_text() == "ping"
            await ws.send_text("pong")

    with client.websocket_connect("/chat") as ws:
        ws.send_text("ping")
        assert ws.receive_text() == "pong"


def test_manual_accept(app: App, client):
    @app.websocket_route("/chat", auto_accept=False)
    async def chat(ws: WebSocket):
        async with ws:
            assert await ws.receive_text() == "ping"
            await ws.send_text("pong")

    with client.websocket_connect("chat") as ws:
        ws.send_text("ping")
        assert ws.receive_text() == "pong"


def test_websocket_route_parameters(app: App, client):
    @app.websocket_route("/chat/{room}")
    async def chat_room(ws: WebSocket, room: str):
        await ws.send(room)

    with client.websocket_connect("/chat/foo") as ws:
        assert ws.receive_text() == "foo"


def test_if_route_parameter_fails_validation_then_403(app: App, client):
    @app.websocket_route("/chat/{id:d}")
    async def chat_room(ws: WebSocket, id: int):
        pass

    with pytest.raises(WebSocketDisconnect) as ctx:
        with client.websocket_connect("/chat/foo"):
            pass
    assert ctx.value.code == 403


def test_non_existing_endpoint_returns_403_as_per_the_asgi_spec(client):
    with pytest.raises(WebSocketDisconnect) as ctx:
        with client.websocket_connect("/foo"):
            pass
    assert ctx.value.code == 403


def test_reject_closes_with_403(app: App, client):
    @app.websocket_route("/foo", auto_accept=False)
    async def foo(ws: WebSocket):
        await ws.reject()

    with pytest.raises(WebSocketDisconnect) as ctx:
        with client.websocket_connect("/foo"):
            pass

    assert ctx.value.code == 403


def test_iter_websocket(app: App, client):
    @app.websocket_route("/chat")
    async def chat(ws: WebSocket):
        async for message in ws:
            await ws.send_text(f"You said: {message}")

    with client.websocket_connect("/chat") as ws:
        ws.send_text("ping")
        assert ws.receive_text() == "You said: ping"
        ws.send_text("pong")
        assert ws.receive_text() == "You said: pong"


@pytest.mark.parametrize("manual", (True, False))
def test_can_close_within_context(app: App, client, manual: bool):
    @app.websocket_route("/test", auto_accept=not manual)
    async def test(ws: WebSocket):
        if manual:
            async with ws:
                await ws.close(4242)
        else:
            await ws.close(4242)

    with client.websocket_connect("/test") as ws:
        message = ws.receive()

    assert message == {"type": "websocket.close", "code": 4242}


def test_websocket_url(app: App, client):
    @app.websocket_route("/test")
    async def test(ws: WebSocket):
        assert ws.url == "ws://testserver/test"
        assert ws.url.path == "/test"
        assert ws.url.port is None
        assert ws.url.scheme == "ws"
        assert ws.url.hostname == "testserver"
        assert ws.url.query == ""
        assert ws.url.is_secure is False

    with client.websocket_connect("/test"):
        pass


def test_websocket_headers(app: App):
    @app.websocket_route("/test")
    async def test(ws: WebSocket):
        await ws.send_json(dict(ws.headers))

    with app.client.websocket_connect("/test", headers={"X-Foo": "bar"}) as ws:
        headers = ws.receive_json()

    assert headers["x-foo"] == "bar"


def test_websocket_query_params(app: App):
    @app.websocket_route("/test")
    async def test(ws: WebSocket):
        await ws.send_json(dict(ws.query_params))

    with app.client.websocket_connect("/test?q=hello") as ws:
        query_params = ws.receive_json()

    assert query_params["q"] == "hello"


# Encoding / decoding of messages


@pytest.mark.parametrize(
    "receive_type, example_message, expected_type",
    [
        ("bytes", b"Hello", bytes),
        ("text", "Hello", str),
        ("json", {"message": "Hello"}, dict),
    ],
)
def test_receive_type(
    app: App, client, receive_type, example_message, expected_type
):
    @app.websocket_route("/chat", receive_type=receive_type)
    async def chat(ws: WebSocket):
        message = await ws.receive()
        assert type(message) == expected_type

    with client.websocket_connect("/chat") as ws:
        getattr(ws, f"send_{receive_type}")(example_message)


@pytest.mark.parametrize(
    "send_type, example_message, expected_type",
    [
        ("bytes", b"Hello", bytes),
        ("text", "Hello", str),
        ("json", {"message": "Hello"}, dict),
    ],
)
def test_send_type(app: App, client, send_type, example_message, expected_type):
    @app.websocket_route("/chat", send_type=send_type)
    async def chat(ws: WebSocket):
        await ws.send(example_message)

    with client.websocket_connect("/chat") as ws:
        message = getattr(ws, f"receive_{send_type}")()
        assert type(message) == expected_type
        assert message == example_message


@pytest.mark.parametrize(
    "value_type, example_message, expected_type",
    [
        ("bytes", b"Hello", bytes),
        ("text", "Hello", str),
        ("json", {"message": "Hello"}, dict),
    ],
)
def test_value_type(
    app: App, client, value_type, example_message, expected_type
):
    @app.websocket_route("/chat", value_type=value_type)
    async def chat(ws: WebSocket):
        message = await ws.receive()
        assert type(message) == expected_type
        await ws.send(example_message)

    with client.websocket_connect("/chat") as ws:
        getattr(ws, f"send_{value_type}")(example_message)
        assert type(getattr(ws, f"receive_{value_type}")()) == expected_type


def test_receive_and_send_event(app: App, client):
    @app.websocket_route("/chat", value_type="event")
    async def chat(ws: WebSocket):
        message = await ws.receive()
        assert message == {"type": "websocket.receive", "text": "ping"}
        await ws.send({"type": "websocket.send", "text": "pong"})

    with client.websocket_connect("/chat") as ws:
        ws.send_text("ping")
        assert ws.receive_text() == "pong"


# Disconnect errors


@pytest.mark.parametrize(
    "close_codes, code, expected_caught",
    [
        *((None, code, True) for code in (1000, 1001)),
        *(
            (None, code, False)
            for code in WEBSOCKET_CLOSE_CODES
            if code not in (1000, 1001)
        ),
        ((1000,), 1001, False),
        ((1000,), 1000, True),
        *(((), code, False) for code in WEBSOCKET_CLOSE_CODES),
        *((all, code, True) for code in WEBSOCKET_CLOSE_CODES),
    ],
)
def test_catch_disconnect(app: App, client, close_codes, code, expected_caught):
    caught = False

    @app.websocket_route(
        "/chat", auto_accept=False, caught_close_codes=close_codes
    )
    async def chat(ws: WebSocket):
        nonlocal caught
        try:
            async with ws:
                await ws.receive()  # will never receive
            caught = True
        except WebSocketDisconnect as exc:
            # The exception should have been raised only if we told the
            # WebSocket route not to catch it.
            assert exc.code not in ws.caught_close_codes

    with client.websocket_connect("/chat") as ws:
        # Close immediately.
        ws.close(code)

    assert caught is expected_caught


# Server error handling


class Oops(Exception):
    pass


def test_if_exception_raised_in_context_then_closed_with_1011(app: App, client):
    @app.websocket_route("/fail")
    async def fail(ws: WebSocket):
        raise Oops

    with suppress(Oops):
        with client.websocket_connect("/fail") as ws:
            message = ws.receive()

    assert message == {"type": "websocket.close", "code": 1011}


def test_accepted_and_exception_raised_then_closed_with_1011(app: App, client):
    @app.websocket_route("/fail", auto_accept=False)
    async def fail(ws: WebSocket):
        await ws.accept()
        raise Oops

    with suppress(Oops):
        with client.websocket_connect("/fail") as ws:
            message = ws.receive()

    assert message == {"type": "websocket.close", "code": 1011}


def test_if_not_accepted_and_exception_raised_then_closed_with_1011(
    app: App, client
):
    @app.websocket_route("/fail", auto_accept=False)
    async def fail(_):
        raise Oops

    with pytest.raises(WebSocketDisconnect) as ctx:
        with client.websocket_connect("/fail"):
            pass

    assert ctx.value.code == 1011


def test_context_does_not_silence_exceptions(app: App, client):
    cleaned_up = False

    @app.websocket_route("/fail", auto_accept=False)
    async def fail(ws):
        nonlocal cleaned_up
        async with ws:
            raise Oops
        cleaned_up = True

    with suppress(Oops):
        with client.websocket_connect("/fail"):
            pass

    assert not cleaned_up



from starlette.responses import RedirectResponse


class Redirection(Exception):
    """Raised when redirecting to another page."""

    def __init__(self, url: str, permanent: bool):
        self._url = url
        self._permanent = permanent

    @property
    def status_code(self) -> int:
        return 301 if self._permanent else 302

    @property
    def response(self) -> RedirectResponse:
        return RedirectResponse(url=self._url, status_code=self.status_code)



import os

import pytest

from bocadillo import App
from bocadillo.utils import override_env


@pytest.fixture(name="empty_run")
def fixture_empty_run():
    def _run(*args, **kwargs):
        pass

    return _run


def test_if_port_env_var_is_set_then_host_is_any_and_port_is_env_var(app: App):
    with override_env("PORT", "4242"):

        def run(app_obj, host, port, **kwargs):
            assert host == "0.0.0.0"
            assert port == 4242
            assert app_obj == app

        app.run(_run=run)


def test_if_port_env_var_is_set_then_specified_host_is_used(app: App):
    with override_env("PORT", "4242"):

        def run(_, host, **kwargs):
            assert host == "example.com"

        app.run(_run=run, host="example.com")


def test_host_is_localhost_by_default(app: App):
    def run(_, host, **kwargs):
        assert host == "127.0.0.1"

    app.run(_run=run)


def test_port_is_8000_by_default(app: App):
    def run(_, port, **kwargs):
        assert port == 8000

    app.run(_run=run)


def test_debug_mode_disabled_by_default(app: App, empty_run):
    assert not app.debug
    app.run(_run=empty_run)
    assert not app.debug


def test_if_debug_then_debug_mode_enabled(app: App, empty_run):
    assert not app.debug
    app.run(debug=True, _run=empty_run)
    assert app.debug


def test_if_debug_env_var_set_then_debug_mode_enabled(app: App, empty_run):
    os.environ["BOCADILLO_DEBUG"] = "1"
    try:
        assert not app.debug
        app.run(_run=empty_run)
        assert app.debug
    finally:
        os.environ.pop("BOCADILLO_DEBUG")


def test_if_debug_then_app_given_as_import_string(app: App):
    def run(app_obj, **kwargs):
        assert isinstance(app_obj, str)
        assert app_obj.endswith(":app")

    app.run(debug=True, _run=run)


def test_if_debug_then_static_files_auto_refresh(app: App, empty_run):
    app.run(debug=True, _run=empty_run)
    for static_app in app._static_apps.values():
        assert static_app.autorefresh


def test_declared_as(app: App):
    def run(app_obj, **kwargs):
        assert app_obj.endswith(":application")

    app.run(debug=True, declared_as="application", _run=run)


def test_if_import_string_unknown_then_no_debug_warning_raised(app: App):
    def run(app, **kwargs):
        pass

    app.import_string = None
    with pytest.warns(UserWarning):
        app.run(debug=True, _run=run)



import pytest

from bocadillo import App, provider, useprovider


@pytest.mark.parametrize("by_name", (False, True))
def test_useprovider(app: App, by_name):
    called = False

    @provider
    async def set_called():
        nonlocal called
        called = True

    @app.route("/")
    @useprovider("set_called" if by_name else set_called)
    async def index(req, res):
        pass

    r = app.client.get("/")
    assert r.status_code == 200
    assert called



import asyncio
import re
from typing import (
    Callable,
    cast,
    List,
    TypeVar,
    Union,
    Optional,
    Any,
    Awaitable,
)

try:
    # >= 3.7
    from contextlib import nullcontext  # pylint: disable=unused-import
except ImportError:  # pragma: no cover
    from contextlib import contextmanager

    @contextmanager
    def nullcontext(enter_result: Any = None):
        yield enter_result


from starlette.concurrency import run_in_threadpool

_CAMEL_REGEX = re.compile(r"(.)([A-Z][a-z]+)")
_SNAKE_REGEX = re.compile(r"([a-z0-9])([A-Z])")

_V = TypeVar("_V")


class asyncnullcontext:
    def __init__(self, enter_result: Any = None):
        self._result = enter_result

    async def __aenter__(self):
        return self._result

    async def __aexit__(self, *args):
        pass


async def call_async(
    func: Union[Callable[..., _V], Callable[..., Awaitable[_V]]],
    *args: Any,
    sync: Optional[bool] = None,
    **kwargs: Any
) -> _V:
    """Call a function in an async manner.

    # Parameters
    func:
        a callable that is either awaited (if a coroutine function)
        or run in the thread pool (if a regular function).
    sync (bool):
        a hint as to whether `func` is synchronous. If not given, it is
        inferred as `asyncio.iscoroutinefunction(func)`.

    # See Also
    - [Executing code in thread or process pools](https://docs.python.org/3/library/asyncio-eventloop.html#executing-code-in-thread-or-process-pools)
    """
    if sync or (sync is None and not asyncio.iscoroutinefunction(func)):
        return await run_in_threadpool(func, *args, **kwargs)

    async_func = cast(Callable[..., Awaitable[_V]], func)
    return await async_func(*args, **kwargs)


def camel_to_snake(name: str) -> str:
    """Convert a `CamelCase` name to its `snake_case` version."""
    s1 = _CAMEL_REGEX.sub(r"\1_\2", name)
    return _SNAKE_REGEX.sub(r"\1_\2", s1).lower()


# WSGI

Environ = dict
StartResponse = Callable[[str, List[str]], None]
WSGIApp = Callable[[Environ, StartResponse], List[bytes]]


def empty_wsgi_app() -> WSGIApp:
    """Return a WSGI app that always returns `404 Not Found`."""

    def wsgi(environ, start_response):
        status = "404 Not Found"
        body = b"Not Found"
        headers = [("Content-Type", "text/plain")]
        start_response(status, headers)
        return [body]

    return wsgi



"""Package setup."""

import setuptools

description = "A modern Python web framework filled with asynchronous salsa."

with open("README.md", "r") as readme:
    long_description = readme.read()

GITHUB = "https://github.com/bocadilloproject/bocadillo"
DOCS = "https://bocadilloproject.github.io"
CHANGELOG = f"{GITHUB}/blob/master/CHANGELOG.md"

INSTALL_REQUIRES = [
    "starlette>=0.11, <0.12",
    "uvicorn>=0.5.1, <0.6",
    "jinja2>=2.10.1",
    "whitenoise",
    "requests",
    "parse",
    "python-multipart",
    "websockets>=6.0",
    "aiodine>=1.2.1, <2.0",
]

EXTRAS_REQUIRE = {"files": ["aiofiles"], "sessions": ["itsdangerous"]}
EXTRAS_REQUIRE["full"] = [
    req for reqs in EXTRAS_REQUIRE.values() for req in reqs
]

setuptools.setup(
    name="bocadillo",
    version="0.13.3",
    author="Florimond Manca",
    author_email="florimond.manca@gmail.com",
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    package_data={"bocadillo": ["assets/*"]},
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    python_requires=">=3.6",
    url=DOCS,
    project_urls={
        "Source": GITHUB,
        "Documentation": DOCS,
        "Changelog": CHANGELOG,
    },
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
    ],
)



from bocadillo import App, hooks
from .utils import function_hooks, async_function_hooks, class_hooks


def test_can_use_function_hooks(app: App, client):
    with function_hooks() as (before, after):

        @app.route("/foo")
        @hooks.before(before)
        @hooks.after(after)
        async def foo(req, res):
            pass

        client.get("/foo")


def test_use_hook_on_sync_function_view(app: App, client):
    with function_hooks() as (before, after):

        @app.route("/foo")
        @hooks.before(before)
        @hooks.after(after)
        def foo(req, res):
            pass

        client.get("/foo")


def test_can_pass_extra_args(app: App, client):
    with function_hooks(after_value=1) as (before, after):

        @app.route("/foo")
        class Foo:
            @hooks.before(before, True)  # positional
            @hooks.after(after, value=1)  # keyword
            async def get(self, req, res):
                pass

        client.get("/foo")


def test_hook_can_be_callable_class(app: App, client):
    with class_hooks() as (before, after):

        @app.route("/foo")
        class Foo:
            @hooks.before(before)
            @hooks.after(after)
            async def get(self, req, res):
                pass

        client.get("/foo")


def test_use_hook_on_view_class(app: App, client):
    with class_hooks() as (before, after):

        @app.route("/foo")
        @hooks.before(before)
        @hooks.after(after)
        class Foo:
            async def get(self, req, res):
                pass

        client.get("/foo")


def test_use_hook_on_method(app: App, client):
    with class_hooks() as (before, after):

        @app.route("/foo")
        class Foo:
            @hooks.before(before)
            @hooks.after(after)
            async def get(self, req, res):
                pass

        client.get("/foo")


def test_use_hook_on_sync_method(app: App, client):
    with class_hooks() as (before, after):

        @app.route("/foo")
        class Foo:
            @hooks.before(before)
            @hooks.after(after)
            def get(self, req, res):
                pass

        client.get("/foo")


def test_hooks_can_be_async(app: App, client):
    with async_function_hooks() as (before, after):

        @app.route("/foo")
        @hooks.before(before)
        @hooks.after(after)
        async def foo(req, res):
            pass

        client.get("/foo")


def test_before_does_not_run_if_method_not_allowed(app: App, client):
    with async_function_hooks(False, False) as (before, after):

        @app.route("/foo")
        @hooks.before(before)
        @hooks.after(after)
        class Foo:
            async def get(self, req, res):
                pass

        response = client.put("/foo")
        assert response.status_code == 405



import inspect
from functools import partial, wraps
from typing import Any, cast, Dict, List, Optional, Type, Union

from . import injection
from .app_types import AsyncHandler, Handler
from .compat import call_async, camel_to_snake
from .constants import ALL_HTTP_METHODS

MethodsParam = Union[List[str], all]  # type: ignore


class HandlerDoesNotExist(Exception):
    # Raised to signal that no handler exists for a requested HTTP method.
    pass


class View:
    """This class defines how all HTTP views are represented internally.

    ::: warning
    Views objects should not be created directly. Instead, use
    [from_handler](#from-handler) or [from_obj](#from-obj).
    :::

    HTTP methods are mapped to **handlers**, i.e. methods of a `View` object.

    The following handlers are supported:

    - `.get(req, res, **kwargs)`
    - `.post(req, res, **kwargs)`
    - `.put(req, res, **kwargs)`
    - `.patch(req, res, **kwargs)`
    - `.delete(req, res, **kwargs)`
    - `.head(req, res, **kwargs)`
    - `.options(req, res, **kwargs)`
    - `.handle(req, res, **kwargs)`

    ::: tip
    `.handle()` is special: if defined, it overrides all others.
    :::

    # Attributes
    name (str): the name of the view.
    """

    def __init__(self, name: str, doc: str = None):
        self.name = name
        if doc is not None:
            self.__doc__ = doc

    get: AsyncHandler
    post: AsyncHandler
    put: AsyncHandler
    patch: AsyncHandler
    delete: AsyncHandler
    head: AsyncHandler
    options: AsyncHandler
    handle: AsyncHandler

    @staticmethod
    def _to_all_async(handlers: Dict[str, Handler]) -> Dict[str, AsyncHandler]:
        async_handlers: Dict[str, AsyncHandler] = {}

        for method, handler in handlers.items():
            if not inspect.iscoroutinefunction(handler):
                handler = wraps(handler)(partial(call_async, handler))
            async_handlers[method] = cast(AsyncHandler, handler)

        return async_handlers

    @classmethod
    def create(
        cls: Type["View"],
        name: str,
        docstring: Optional[str],
        handlers: Dict[str, Handler],
    ) -> "View":
        async_handlers: Dict[str, AsyncHandler] = cls._to_all_async(handlers)

        copy_get_to_head = (
            "get" in async_handlers and "head" not in async_handlers
        )
        if copy_get_to_head:
            async_handlers["head"] = async_handlers["get"]

        vue: View = cls(name, doc=docstring)

        for method, handler in async_handlers.items():
            handler = injection.consumer(handler)
            setattr(vue, method, handler)

        return vue

    def get_handler(self, method: str) -> AsyncHandler:
        try:
            return getattr(self, "handle")
        except AttributeError:
            try:
                return getattr(self, method)
            except AttributeError as exc:
                raise HandlerDoesNotExist from exc


def from_handler(handler: Handler, methods: MethodsParam = None) -> View:
    """Convert a handler to a #::bocadillo.views#View instance.

    # Parameters
    handler (function or coroutine function):
        Its name and docstring are copied onto the view.
        It used as a handler for each of the declared `methods`.
        For example, if `methods=["post"]` then the returned `view` object
        will only have a `.post()` handler.
    methods (list of str):
        A list of supported HTTP methods. The `all` built-in can be used
        to support all HTTP methods. Defaults to `["get"]`.

    # Returns
    view: a #::bocadillo.views#View instance.

    # See Also
    - The [constants](./constants.md) module for the list of all HTTP methods.
    """
    if methods is None:
        methods = ["get"]
    if methods is all:
        methods = ["handle"]
    else:
        methods = [m.lower() for m in methods]
    handlers = {method: handler for method in methods}
    return View.create(handler.__name__, handler.__doc__, handlers)


def from_obj(obj: Any) -> View:
    """Convert an object to a #::bocadillo.views#View instance.

    # Parameters
    obj (any):
        its handlers, snake-cased class name and docstring are copied
        onto the view.

    # Returns
    view: a #::bocadillo.views#View instance.
    """
    handlers = get_handlers(obj)
    name = camel_to_snake(obj.__class__.__name__)
    return View.create(name, obj.__doc__, handlers)


def get_handlers(obj: Any) -> Dict[str, Handler]:
    """Return all #::bocadillo.views#View handlers declared on an object.

    # Parameters
    obj (any): an object.

    # Returns
    handlers (dict):
        A dict mapping an HTTP method to a handler.
    """
    all_methods = map(str.lower, ALL_HTTP_METHODS)
    try:
        handle = obj.handle
    except AttributeError:
        return {
            method: getattr(obj, method)
            for method in all_methods
            if hasattr(obj, method)
        }
    else:
        return {method: handle for method in all_methods}


def view(methods: MethodsParam = None):
    """Convert the decorated function to a proper #::bocadillo.views#View.

    This decorator is a shortcut for [from_handler](#from-handler).
    """
    return partial(from_handler, methods=methods)
